﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;

using NDesk.Options;

namespace rbcom {
    public abstract class Command {

        protected const int RETURN_OK = 0;
        protected const int RETURN_FAILED = 1;
        protected const int RETURN_ERROR = -1;
        protected const int RETURN_INVALID = -2;

        /// <summary>
        /// コマンド名
        /// </summary>
        public string Name {
            get {
                return Path.GetFileNameWithoutExtension(Assembly.GetEntryAssembly().Location);
            }
            private set {
                // DO NOTHING
            }
        }

        /// <summary>
        /// コマンドの詳細説明
        /// </summary>
        public string Description { get; protected set; }

        /// <summary>
        /// コマンド説明のフッタ
        /// </summary>
        protected string Footer { get; set; }

        /// <summary>
        /// 基本利用方法
        /// </summary>
        protected string Usage { get; set; }

        /// <summary>
        /// Ruby Manager
        /// </summary>
        protected RubyProgramManager Rubies { get; private set; }

        /// <summary>
        /// オプションリスト
        /// </summary>
        protected NDesk.Options.OptionSet Options { get; private set; }

        /// <summary>
        /// 冗長モードフラグ
        /// </summary>
        protected bool VerboseMode { get; set; }

        /// <summary>
        /// Usage表示フラグ
        /// </summary>
        protected bool UsageMode { get; set; }

        /// <summary>
        /// 設定ファイルパス
        /// </summary>
        protected string ConfigurationPath { get; set; }

        /// <summary>
        /// 設定ファイル名
        /// </summary>
        protected string ConfigurationName { get; set; }

        /// <summary>
        /// 設定ファイルの初期値
        /// </summary>
        protected string DefaultConfigurationPath {
            get {
                return Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), this.ConfigurationName);
            }
            private set {
                // DO NOTHING
            }
        }

        protected bool NoLoadConfiguration { get; set; }
        protected bool NoSaveConfiguration { get; set; }

        public Command(string desc = "", string footer = "") {
            Debug.Listeners.Add(new TextWriterTraceListener(System.Console.Error));

            // 基本設定
            this.ConfigurationName = "rbcom.json";
            this.Description = desc;
            this.Footer = footer;
            this.NoLoadConfiguration = false;
            this.NoSaveConfiguration = false;

            // Build Options
            this.Options = new OptionSet();
            this.Options.Add("verbose|V", "冗長出力を設定します", delegate(string v) { this.VerboseMode = true; });
            this.Options.Add("help|h", "ヘルプメッセージを表示します", delegate(string v) { this.UsageMode = true; });
            this.Options.Add("config=", "使用する設定ファイル", delegate(string v) { this.ConfigurationPath = v; });
            this.Rubies = new RubyProgramManager();
        }

        public virtual void printUsage(System.IO.TextWriter writer) {
            if (this.Usage == null) {
                writer.WriteLine(String.Format("USAGE: {0} [options]", this.Name));
            } else {
                writer.WriteLine(this.Usage);
            }
            writer.WriteLine();

            if (this.Description != null && this.Description.Length > 0) {
                writer.WriteLine(this.Description);
                writer.WriteLine();
            }

            // オプションを表示
            if (this.Options.Count > 0) {
                writer.WriteLine("オプションパラメータ:");
                this.Options.WriteOptionDescriptions(writer);
            } else {
                writer.WriteLine("オプションパラメータはありません");
            }

            // フッタ
            if (this.Footer != null && this.Footer.Length > 0) {
                writer.WriteLine();
                writer.WriteLine(this.Footer);
                writer.WriteLine();
            }
        }

        public int Execute(string[] args) {
            List<string> argv = this.PreInvoke(args);

            if (this.UsageMode) {
                this.printUsage(System.Console.Out);
                return RETURN_OK;
            }
            if (!this.Validate(argv.ToArray())) {
                this.printUsage(System.Console.Error);
                return RETURN_INVALID;
            }

            if (this.ConfigurationPath == null) {
                this.ConfigurationPath = this.DefaultConfigurationPath;
            }

            int ret = RETURN_OK;
            try {
                // Load Configuration
                this.LoadConfiguration();

                // Execute
                if (this.Invoke(argv.ToArray())) {
                    ret = RETURN_OK;
                } else {
                    ret = RETURN_FAILED;
                }

                // Save Configuration
                this.SaveConfiguration();
            } catch (Exception e) {
                System.Console.Error.WriteLine(e.ToString());
                ret = RETURN_ERROR;
            } finally {
                this.PostInvoke();
            }

            return ret;
        }

        protected virtual List<string> PreInvoke(string[] args) {
            return this.Options.Parse(args);
        }

        protected virtual bool Validate(string[] args) {
            return true;
        }

        protected abstract bool Invoke(string[] args);

        protected virtual void PostInvoke() {
            // Implemented in sub-classes if needed.
        }

        /// <summary>
        /// ConfigurationPathから設定を読み込みます
        /// </summary>
        protected void LoadConfiguration() {
            if (this.NoLoadConfiguration) {
                if (this.VerboseMode) {
                    System.Console.Error.WriteLine("NOTICE: Skipped to load configuration");
                }
            } else {
                // Loading Configuration
                if (this.VerboseMode) {
                    System.Console.Error.Write(String.Format("INFO: Loading Configuration from {0}...", this.ConfigurationPath));
                }
                if (this.Rubies.Load(this.ConfigurationPath)) {
                    if (this.VerboseMode) {
                        System.Console.Error.WriteLine("OK");
                    }
                } else {
                    if (this.VerboseMode) {
                        System.Console.Error.WriteLine("Failed");
                    }
                }
            }
        }

        /// <summary>
        /// ConfigurationPathへ設定を書き込みます
        /// </summary>
        protected void SaveConfiguration() {
            if (this.NoSaveConfiguration) {
                if (this.VerboseMode) {
                    System.Console.Error.WriteLine("NOTICE: Skipped to save configuration");
                }
            } else {
                // Saving Configuration
                if (this.VerboseMode) {
                    System.Console.Error.Write("INFO: Saving Configuration into {0}...", this.ConfigurationPath);
                }
                if (this.Rubies.Save(this.ConfigurationPath)) {
                    if (this.VerboseMode) {
                        System.Console.Error.WriteLine("OK");
                    }
                } else {
                    if (this.VerboseMode) {
                        System.Console.Error.WriteLine("Failed");
                    }
                }
            }
        }

        /// <summary>
        /// コマンドを実行します.
        /// OSにはExecuteが返した戻り値が返されます
        /// </summary>
        /// <param name="command">対象コマンド</param>
        /// <param name="args">コマンド引数</param>
        public static void Run(Command command, string[] args) {
            Environment.Exit(command.Execute(args));
        }

    }
}
