﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

using System.Runtime.Serialization;


namespace rbcom {
    [DataContractAttribute]
    public class RubyVersion : IComparable {

        [DataMemberAttribute(EmitDefaultValue = true, IsRequired = true, Name = @"major", Order = 0)]
        public int Major { get; set; }

        [DataMemberAttribute(EmitDefaultValue = true, IsRequired = true, Name = @"minor", Order = 1)]
        public int? Minor { get; set; }

        [DataMemberAttribute(EmitDefaultValue = true, IsRequired = true, Name = @"teeny", Order = 2)]
        public int? Teeny { get; set; }

        [DataMemberAttribute(EmitDefaultValue = true, IsRequired = false, Name = @"patchlevel", Order = 3)]
        public int? PatchLevel { get; set; }

        /// <summary>
        /// バージョンの一致を調べます
        /// </summary>
        /// <param name="left"></param>
        /// <param name="right"></param>
        /// <returns></returns>
        public static bool operator ==(RubyVersion left, RubyVersion right) {
            if (((object) left) == null && ((object) right) == null) {
                return true;
            } else if (((object) left) == null || ((object) right) == null) {
                return false;
            }

            int lmajor = left.Major;
            int lminor = (left.Minor.HasValue ? left.Minor.Value : 0);
            int lteeny = (left.Teeny.HasValue ? left.Teeny.Value : 0);
            int lpatch = (left.PatchLevel.HasValue ? left.PatchLevel.Value : 0);
            int rmajor = right.Major;
            int rminor = (right.Minor.HasValue ? right.Minor.Value : 0);
            int rteeny = (right.Teeny.HasValue ? right.Teeny.Value : 0);
            int rpatch = (right.PatchLevel.HasValue ? right.PatchLevel.Value : 0);

            return ((lmajor == rmajor) && (lminor == rminor) && (lteeny == rteeny) && (lpatch == rpatch));
        }

        /// <summary>
        /// バージョンの不一致を調べます
        /// </summary>
        /// <param name="left"></param>
        /// <param name="right"></param>
        /// <returns></returns>
        public static bool operator !=(RubyVersion left, RubyVersion right) {
            return !(left == right);
        }

        /// <summary>
        /// バージョンを比較し、左側が右側より大きいかを調べます
        /// </summary>
        /// <param name="left"></param>
        /// <param name="right"></param>
        /// <returns></returns>
        public static bool operator >(RubyVersion left, RubyVersion right) {
            if (((object)left) == null && ((object)right) == null) {
                return true;
            } else if (((object)left) == null || ((object)right) == null) {
                return false;
            }

            int lmajor = left.Major;
            int lminor = (left.Minor.HasValue ? left.Minor.Value : 0);
            int lteeny = (left.Teeny.HasValue ? left.Teeny.Value : 0);
            int lpatch = (left.PatchLevel.HasValue ? left.PatchLevel.Value : 0);
            int rmajor = right.Major;
            int rminor = (right.Minor.HasValue ? right.Minor.Value : 0);
            int rteeny = (right.Teeny.HasValue ? right.Teeny.Value : 0);
            int rpatch = (right.PatchLevel.HasValue ? right.PatchLevel.Value : 0);

            if (lmajor < rmajor) {
                return false;
            } else if (lmajor > rmajor) {
                return true;
            }
            if (lminor < rminor) {
                return false;
            } else if (lminor > rminor) {
                return true;
            }
            if (lteeny < rteeny) {
                return false;
            } else if (lteeny > rteeny) {
                return true;
            }
            if (lpatch > rpatch) {
                return true;
            } else {
                return false;
            }
        }

        /// <summary>
        /// 左右のバージョンが同じか、左側大きいかどうかを調べます
        /// </summary>
        /// <param name="left"></param>
        /// <param name="right"></param>
        /// <returns></returns>
        public static bool operator >=(RubyVersion left, RubyVersion right) {
            return (left == right) || (left > right);
        }

        /// <summary>
        /// 左右のバージョンが同じか、右側のほうが大きいかを調べます
        /// </summary>
        /// <param name="left"></param>
        /// <param name="right"></param>
        /// <returns></returns>
        public static bool operator <=(RubyVersion left, RubyVersion right) {
            return !(left > right);
        }

        /// <summary>
        /// バージョンを比較し、右側のほうが大きいかどうかを調べます
        /// </summary>
        /// <param name="left"></param>
        /// <param name="right"></param>
        /// <returns></returns>
        public static bool operator <(RubyVersion left, RubyVersion right) {
            return !(left >= right);
        }

        /// <summary>
        /// バージョン番号からバージョンオブジェクトを作成します
        /// </summary>
        /// <param name="major">メジャーバージョン番号</param>
        /// <param name="minor">マイナーバージョン番号</param>
        /// <param name="teeny">なんかのバージョン番号</param>
        /// <param name="patch">パッチレベル</param>
        public RubyVersion(int major, int? minor = null, int? teeny = null, int? patch = null) {
            this.Major = major;
            this.Minor = minor;
            this.Teeny = teeny;
            this.PatchLevel = patch;
        }

        /// <summary>
        /// バージョン文字列からバージョンオブジェクトを作成します
        /// </summary>
        /// <param name="version">バージョン文字列。1.9.3-p371 形式</param>
        public RubyVersion(string version) {
            string pattern = @"(?<major>\d+)(\.(?<minor>\d+)(\.(?<teeny>\d+))?)?(\-?[pP](?<patch>\d+))?";
            Match match = Regex.Match(version, pattern);

            // 解析
            int major = 0;
            int? minor = null;
            int? teeny = null;
            int? patch = null;
            if (match.Groups[@"major"].Success) {
                Int32.TryParse(match.Groups[@"major"].Value, out major);
            }
            if (match.Groups[@"minor"].Success) {
                int minorValue = 0;
                Int32.TryParse(match.Groups[@"minor"].Value, out minorValue);
                minor = minorValue;
            }
            if (match.Groups[@"teeny"].Success) {
                int teenyValue = 0;
                Int32.TryParse(match.Groups[@"teeny"].Value, out teenyValue);
                teeny = teenyValue;
            }
            if (match.Groups[@"patch"].Success) {
                int patchLevel = 0;
                Int32.TryParse(match.Groups[@"patch"].Value, out patchLevel);
                patch = patchLevel;
            }

            // 設定
            this.Major = major;
            this.Minor = minor;
            this.Teeny = teeny;
            this.PatchLevel = patch;
        }

        public bool Match(RubyVersion other) {
            bool majorMatch = (this.Major == other.Major);
            bool minorMatch = false;
            bool teenyMatch = false;
            bool patchMatch = false;
            if (!this.Minor.HasValue || !other.Minor.HasValue) {
                minorMatch = true;
            } else {
                minorMatch = (this.Minor == other.Minor);
            }
            if (!this.Teeny.HasValue || !other.Teeny.HasValue) {
                teenyMatch = true;
            } else {
                teenyMatch = (this.Teeny == other.Teeny);
            }
            if (!this.PatchLevel.HasValue || !other.PatchLevel.HasValue) {
                patchMatch = true;
            } else {
                patchMatch =  (this.PatchLevel == other.PatchLevel);
            }

            return majorMatch && minorMatch && teenyMatch && patchMatch;
        }

        public int CompareTo(object obj) {
            if (obj is RubyVersion) {
                RubyVersion other = (RubyVersion)obj;
                int dMajor = (this.Major - other.Major);
                if (dMajor != 0) {
                    return dMajor;
                }

                // Minor Version
                if (this.Minor.HasValue && other.Minor.HasValue) {
                    int dMinor = (this.Minor.Value - other.Minor.Value);
                    if (dMinor != 0) {
                        return dMinor;
                    }
                } else if (this.Minor.HasValue && !other.Minor.HasValue) {
                    return this.Minor.Value;
                } else if (!this.Minor.HasValue && other.Minor.HasValue) {
                    return -other.Minor.Value;
                }

                // Teeny Version
                if (this.Teeny.HasValue && other.Teeny.HasValue) {
                    int dTeeny = (this.Teeny.Value - other.Teeny.Value);
                    if (dTeeny != 0) {
                        return dTeeny;
                    }
                } else if (this.Teeny.HasValue && !other.Teeny.HasValue) {
                    return this.Teeny.Value;
                } else if (!this.Teeny.HasValue && other.Teeny.HasValue) {
                    return -other.Teeny.Value;
                }

                // Patch Level
                if (this.PatchLevel.HasValue && other.PatchLevel.HasValue) {
                    return (this.PatchLevel.Value - other.PatchLevel.Value);
                } else if (this.PatchLevel.HasValue && !other.PatchLevel.HasValue) {
                    return this.PatchLevel.Value;
                } else if (!this.PatchLevel.HasValue && other.PatchLevel.HasValue) {
                    return -other.PatchLevel.Value;
                } else {
                    return 0;
                }
            } else {
                throw new ArgumentException("object is not RubyVersion");
            }
        }

        public override bool Equals(object obj) {
            if (obj.GetType() == (typeof(RubyVersion))) {
                RubyVersion left = this;
                RubyVersion right = (RubyVersion) obj;

                if (left.Major != right.Major) {
                    return false;
                }
                if (left.Minor != right.Minor) {
                    return false;
                }
                if (left.Teeny != right.Teeny) {
                    return false;
                }

                // PatchLevel
                if (left.PatchLevel == null && right.PatchLevel == null) {
                    return true;
                } else if (left.PatchLevel != null && right.PatchLevel != null) {
                    return (left.PatchLevel == right.PatchLevel);
                } else {
                    return false;
                }
            } else {
                return base.Equals(obj);
            }
        }

        public override int GetHashCode() {
            int majorHash = this.Major << 24;
            int minorHash = 0;
            if (this.Minor.HasValue) {
                minorHash = this.Minor.Value << 16;
            }
            int teenyHash = 0;
            if (this.Teeny.HasValue) {
                teenyHash = this.Teeny.Value << 8;
            }
            int hashCode = (majorHash | minorHash | teenyHash);
            if (this.PatchLevel != null) {
                hashCode |= this.PatchLevel.Value;
            }
            return hashCode;
        }

        public override string ToString() {
            string major = string.Format("{0}", this.Major);
            string minor = null;
            string teeny = null;
            if (this.Minor.HasValue) {
                minor = String.Format("{0}", this.Minor);
            } else {
                minor = @"*";
            }
            if (this.Teeny.HasValue) {
                teeny = String.Format("{0}", this.Teeny);
            } else {
                teeny = @"*";
            }
            if (this.PatchLevel == null) {
                return String.Format("{0}.{1}.{2}", major, minor, teeny);
            } else {
                return String.Format("{0}.{1}.{2}-p{3}", major, minor, teeny, this.PatchLevel);
            }
        }

    }
}
